/* Copyright (C) 2023-2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaOutputQa.cxx
/// \date   20.11.2023
/// \brief  A QA module for CA tracking (implementation)
/// \author S.Zharko <s.zharko@gsi.de>

#include "CaOutputQa.h"

#include "CaConstants.h"
#include "CaInputData.h"
#include "CaParameters.h"
#include "CaTrack.h"

#include <fmt/format.h>

using cbm::algo::ca::OutputQa;
using cbm::algo::ca::constants::math::Pi;

// ---------------------------------------------------------------------------------------------------------------------
//
void OutputQa::Init()
{
  using cbm::algo::qa::CanvasConfig;
  using cbm::algo::qa::Data;
  using cbm::algo::qa::H1D;
  using cbm::algo::qa::H2D;
  using cbm::algo::qa::PadConfig;
  using fmt::format;

  if (!fpSender.get()) {
    return;
  }

  // ----- Histograms initialization
  constexpr int nParPads                              = 3;
  std::array<std::string, knTrkParPoints> vsPointName = {"first", "last"};
  for (int i = 0; i < knTrkParPoints; ++i) {
    {
      auto sName      = format("track_{}_theta", vsPointName[i]);
      auto sTitl      = format("#theta at {} hit; #theta", vsPointName[i]);
      fvphTrkTheta[i] = fQaData.MakeObj<H1D>(sName, sTitl, 62, 0., 90.);
    }
    {
      auto sName    = format("track_{}_phi", vsPointName[i]);
      auto sTitl    = format("#phi at {} hit; #phi", vsPointName[i]);
      fvphTrkPhi[i] = fQaData.MakeObj<H1D>(sName, sTitl, 62, -180., 180.);
    }
    {
      auto sName         = format("track_{}_thata_phi", vsPointName[i]);
      auto sTitl         = format("#theta vs #phi at {} hit; #phi; #theta", vsPointName[i]);
      fvphTrkPhiTheta[i] = fQaData.MakeObj<H2D>(sName, sTitl, 62, -180., 180., 62, 0., 90.);
    }
    {
      auto sName        = format("track_{}_chi2_ndf", vsPointName[i]);
      auto sTitl        = format("#chi^{{2}}/NDF at {} hit; #chi^{{2}}/NDF", vsPointName[i]);
      fvphTrkChi2Ndf[i] = fQaData.MakeObj<H1D>(sName, sTitl, 100, 0., 20.);
    }
  }
  {
    int nBins   = knStaMax;
    double xMin = -0.5;
    double xMax = double(knStaMax) - 0.5;
    {
      auto sName      = "track_fst_lst_sta";
      auto sTitl      = "First vs. last station index;ID^{last}_{station};ID^{first}_{station}";
      fphTrkFstLstSta = fQaData.MakeObj<H2D>(sName, sTitl, nBins, xMin, xMax, nBins, xMin, xMax);
    }
    fphTrkNofHits = fQaData.MakeObj<H1D>("n_hits", "Number of hits;N_{hit}", nBins, xMin, xMax);
  }

  // ---- Init canvases
  {

    // Track parameters at first/last track hit

    for (int i = 0; i < knTrkParPoints; ++i) {
      auto sCanvName = fmt::format("ca_trk_{}_hit", vsPointName[i]);
      auto sCanvTitl = fmt::format("Track parameters at {} hit", vsPointName[i]);
      auto canv      = CanvasConfig(sCanvName, sCanvTitl);
      std::vector<PadConfig> pads;
      pads.reserve(nParPads);
      for (int iPad = 0; iPad < nParPads; ++iPad) {
        auto& pad = pads.emplace_back();
        pad.SetLog(false, true);
        pad.SetGrid(true, true);
      }
      pads[0].RegisterHistogram(fvphTrkTheta[i], "hist");
      pads[1].RegisterHistogram(fvphTrkPhi[i], "hist");
      pads[2].RegisterHistogram(fvphTrkChi2Ndf[i], "hist");
      for (auto& pad : pads) {
        canv.AddPadConfig(pad);
      }
      fQaData.AddCanvasConfig(canv);
    }

    {
      auto sCanvName = "ca_trk_theta_phi";
      auto sCanvTitl = "Track #theta vs #phi";
      auto canv      = CanvasConfig(sCanvName, sCanvTitl);
      for (int i = 0; i < knTrkParPoints; ++i) {
        PadConfig pad;
        pad.SetLog(false, false, true);
        pad.SetGrid(true, true);
        pad.RegisterHistogram(fvphTrkPhiTheta[i], "colz");
        canv.AddPadConfig(pad);
      }
      fQaData.AddCanvasConfig(canv);
    }

    // Number of hits in track
    {
      auto canv = CanvasConfig("ca_nhits_in_trk", "Number of hit in track");
      {
        auto pad = PadConfig();
        pad.SetGrid(true, true);
        pad.SetLog(false, true);
        pad.RegisterHistogram(fphTrkNofHits, "hist");
        canv.AddPadConfig(pad);
      }
      {
        auto pad = PadConfig();
        pad.SetGrid(true, true);
        pad.SetLog(false, false, true);
        pad.RegisterHistogram(fphTrkFstLstSta, "colz");
        canv.AddPadConfig(pad);
      }
      fQaData.AddCanvasConfig(canv);
    }
  }

  fQaData.Init(fpSender);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void OutputQa::Exec()
{
  if (!fpSender.get()) {
    return;
  }

  if (!CheckInit()) {
    L_(fatal) << "ca::OutputQa: instance is not initialized";
    assert(false);
  }

  // Fill track histograms
  {
    int trkFirstHit = 0;  // Index of hit in fpvRecoHits
    for (auto trkIt = fpvTracks->begin(); trkIt != fpvTracks->end(); ++trkIt) {
      auto& track = *trkIt;
      int nHits   = track.fNofHits;
      // Indices of hits in fpInputData->GetHits()
      int iFstHit = (*fpvRecoHits)[trkFirstHit];
      int iLstHit = (*fpvRecoHits)[trkFirstHit + nHits - 1];

      // Distributions in different track points
      for (int ip = 0; ip < knTrkParPoints; ++ip) {
        //int iHit           = (ip == 0 ? iFstHit : iLstHit);
        //const auto& hit    = fpInputData->GetHit(iHit);
        const auto& trkPar = (ip == 0 ? track.fParFirst : track.fParLast);
        fvphTrkTheta[ip]->Fill(trkPar.GetTheta() * 180. / Pi);
        fvphTrkPhi[ip]->Fill(trkPar.GetPhi() * 180. / Pi);
        fvphTrkPhiTheta[ip]->Fill(trkPar.GetPhi() * 180. / Pi, trkPar.GetTheta() * 180. / Pi);
        fvphTrkChi2Ndf[ip]->Fill(trkPar.GetChiSq() / trkPar.GetNdf());
      }
      // Other distributions
      fphTrkFstLstSta->Fill(fpInputData->GetHit(iLstHit).Station(), fpInputData->GetHit(iFstHit).Station());
      fphTrkNofHits->Fill(nHits);
      trkFirstHit += nHits;
    }
  }

  fQaData.Send(fpSender);
}
