/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Dominik Smith [committer] */

#ifndef ALGO_DETECTORS_TOF_READOUTCONFIG_H
#define ALGO_DETECTORS_TOF_READOUTCONFIG_H

#include <cstddef>
#include <cstdint>
#include <map>
#include <vector>

namespace cbm::algo::tof::config
{
  struct CROB;
  struct ReadoutPars;
}  // namespace cbm::algo::tof::config

namespace cbm::algo::tof
{

  class ReadoutConfig {

   public:
    /** @brief Constructor **/
    ReadoutConfig(const config::ReadoutPars& pars);

    /** @brief Destructor **/
    ~ReadoutConfig();

    /** @brief Equipment in the configuration
     ** @return Vector of equipment IDs
     **/
    std::vector<uint16_t> GetEquipmentIds();

    /** @brief Number of elinks of a component
     ** @param Equipment ID
     ** @return Number of elinks
     **/
    size_t GetNumElinks(uint16_t equipmentId);

    /** @brief API: Mapping from component and elink to addresses per channel
     ** @param equipId     Equipment identifier (component)
     ** @param elink       Elink number within component
     ** @return Vector of TOF addresses, indexed via channel number
     */
    std::vector<uint32_t> Map(uint16_t equipId, uint16_t elink);

   private:
    // --- TOF readout map
    // --- Map index: (equipment, elink, channel), map value: (TOF address)
    std::map<uint16_t, std::vector<std::vector<uint32_t>>> fReadoutMap = {};

    /** @brief Initialisation of readout map **/
    void Init(const config::ReadoutPars& pars);

    /// Mapping to eLink to ASIC number within DPB. Mapping is the same for each DPB.
    int32_t ElinkIdxToGet4Idx(uint32_t elink, const config::ReadoutPars& pars);
    /// Mapping to eLink to ASIC number within DPB for 2024 inner TOF FEBs. Mapping is the same for each DPB.
    int32_t ElinkIdxToGet4IdxInner(uint32_t elink, const config::ReadoutPars& pars);

    int32_t Get4ChanToPadiChan(uint32_t channelInFee, const config::ReadoutPars& pars);

    std::vector<int32_t> fviRpcChUId = {};  // UID/address for each channel, build from type, side and module

    void BuildChannelsUidMap(const config::ReadoutPars& pars);
    void BuildChannelsUidMapCbm(uint32_t& uCh, const config::CROB& crob);
    void BuildChannelsUidMapStar(uint32_t& uCh, const config::CROB& crob);
    void BuildChannelsUidMapCern(uint32_t& uCh, const config::CROB& crob);
    void BuildChannelsUidMapCera(uint32_t& uCh, const config::CROB& crob);
    void BuildChannelsUidMapStar2(uint32_t& uCh, const config::CROB& crob);
    void BuildChannelsUidMapStar2Inner(uint32_t& uCh, const config::CROB& crob);
    void BuildChannelsUidMapBuc(uint32_t& uCh, const config::CROB& crob);
  };

}  // namespace cbm::algo::tof

#endif  //ALGO_DETECTORS_TOF_READOUTCONFIG_H
